/************************************************************************************
  USBMASCON.ino   S88-N Version
   This sketch can be used by connecting ZUIKI's SWitch mascon via the USB HOST shield.
   The source is created by referring to the sample USB HID Boystick of the USB HOST shield.

   This sketch requires the following libraries:
    USB_Host_Shield_2.0
    I2C 128x64 OLED Driver Library   (used the libraries included in DS HACX R4)
    This scketch is Refered to the sketch of S88_Mascon_K9 by Fujigaya2.

    
    Control sketch based on the S88-N specifications.
    Confirmed the operation with the command station made by DesktopStation.


   2021/09/20 Rev02
   Copyright (c) 2021 Nagoden  All right reserved.
*************************************************************************************/
#include <usbhid.h>
#include <hiduniversal.h>
#include <usbhub.h>
#include <EEPROM.h>

// Satisfy IDE, which only needs to see the include statment in the ino.
#ifdef dobogusinclude
#include <spi4teensy3.h>
#endif
#include <SPI.h>
#include "hidjoystickrptparser.h"

//OLED library
#include "OzOLED.h"

//----------------------------------------------------------------------
//マスコン加減速定数
//ブレーキは減速数　加速は加速数
//0から1023までで速度制御するのでノッチの位置でどれだけ足し引きするかの定数
//CONST_NNは、ニュートラル時の減速度合いで　0の場合は減速しない　
#define CONST_EB 20
#define CONST_B8 15
#define CONST_B7 10
#define CONST_B6 8
#define CONST_B5 5
#define CONST_B4 4
#define CONST_B3 3
#define CONST_B2 2
#define CONST_B1 1
#define CONST_NN 1
#define CONST_P1 10
#define CONST_P2 5
#define CONST_P3 5
#define CONST_P4 5
#define CONST_P5 5
//マスコン速度リミット％
//Pノッチの位置でどこまでを最高とするかの率を％で指定
#define LIM_P1 10
#define LIM_P2 30
#define LIM_P3 50
#define LIM_P4 80
#define LIM_P5 100
//ボタンのファンクション割り当て(F0～F28まで割り当て可能)
//99を指定するとそのボタンのファンクション機能は無効
#define BUTTON_A 2
#define BUTTON_B 3
#define BUTTON_Y 4
#define BUTTON_X 6
#define BUTTON_ZL 5
#define BUTTON_L 0
#define BUTTON_M 1
#define BUTTON_P 7
#define BUTTON_R 8

//----------------------------------------------------------------------
// Settings used in S88N
#define PWR_LED_PIN A1          // PWR_LED(Power LED) on/off pin
#define RUN_LED_PIN A0          // RUN_LED(status LED) on/off pin
#define SENSOR_INTERVAL 5       //sensor interval[ms]
#define CONT_PORT_NUM 16        //port_num to see from DCC Controller
#define EEP_LOCO_NO_DEFAULT 3   //Default Loco Address

//S88
// reference : S88_Mascon_K9
// Fujigaya2 2016/09/17
// https://seesaawiki.jp/fujigaya2/

// reference : RB_S88
// Software by Ruud Boer, November 2014.
// https://rudysmodelrailway.wordpress.com/software/

int clk_counter=0;
const byte dataIn=5;  //data input from next Arduino in S88 chain
const byte dataOut=6; //data output pin=6


/*********************************************************************************
 Variable definition
 ********************************************************************************/

//走行用
int Direction = 0;  //0でFWD  1でRWD
long gSpeed = 0;    //SPEED MAX1023
int SpeedRate = 0;  //スピード比率（バー表示のための0-255の値）　
int gMspeed = 0;    //スピード値保持用     
//ACC制御
uint16_t acc_adr = 0; //アクセサリ用アドレス 範囲は0-2044
int acc_pos = 0;      //Str,Divフラグ

//OLED表示用文字変数
char gText[16 + 1];
int gLocoAddresses[] = {0, 0, 0, 0};
int gbkLocoAddresses[] = {0, 0, 0, 0};  //アドレス変更時のバックアップ
int gswAddresses[] = {0, 0, 0, 0};
int gbkswAddresses[] = {0, 0, 0, 0};  //アクセサリスイッチアドレス変更時のバックアップ
char* space14 = "              ";

//DCCアドレス
int gdecLocoadddress = 0003 ;   //10進表記でのアドレス
uint16_t glocoadress = 0;       //バイナリでのアドレス
int loco_no = EEP_LOCO_NO_DEFAULT;

//マスコンボタンのイベント発生時の変数
byte buttonBufRz = 0;
byte buttonBufX1 = 0;
byte buttonBufX2 = 0;
byte buttonBufY1 = 0;

//制御用
int ValueFunction[28] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int gTimeState = 0;           //コマンド受信待ちの処理のステータスフラグ
int gThotteError = 0;         //アドレスコマンド送出後のエラー取得用
int gStepmode = 0;            //S88－ではステップモードは未使用

//S88 Control
uint16_t transfer_value;      //出力の変数
uint16_t sensor_value;        //出力の変数
uint16_t out_value;           //出力の変数
boolean out_flag = false;     //出力の変数
boolean Power_Flg = false;    //線路電源　ON/OFF
int set_s88_switch = 4;
//0:速度・進行方向
//1:ファンクション１
//2:ファンクション２
//3:Acc Address
//4:Loc Address
//100:空
int SendReserveFlag = 100;
int loco_dir = 0;             //S88-nで使用するスピードと方向の変数
int loco_speed = 0;           //S88-nで使用するスピードと方向の変数
int loco_dir_prev = 0;        //S88-nで使用するスピードと方向の変数（バックアップ）
int loco_speed_prev = 0;      //S88-nで使用するスピードと方向の変数（バックアップ）

/*********************************************************************************
 Menu definition
 ********************************************************************************/
uint8_t SetAddress_mode;
int wXcursor;                 //メニュー制御ｘアドレスワークカーソル
int wadXcursor;               //メニュー制御ｘアクセサリワークカーソル

/*********************************************************************************
    Task Schedule
 ********************************************************************************/

unsigned long gPreviousL1 = 0; //   300ms interval  Acceleration / deceleration processing after throttle event
unsigned long gPreviousL2 = 0; //   500ms interval  Automatic deceleration in neutral
unsigned long gPreviousL3 = 0; //   150ms interval  S88 Process
/********************************************************************************
   JoystickEventsクラスを継承
 ********************************************************************************/
class MyJoystickEvents : public JoystickEvents {
  public:
    void OnGamePadChanged(const GamePadEventData *evt);
};

USB Usb;
USBHub Hub(&Usb);
HIDUniversal Hid(&Usb);
MyJoystickEvents MyJoyEvents;            //Make it a new inherited class.
JoystickReportParser Joy(&MyJoyEvents);  //Use an instance of a new class

/*********************************************************************************
  Definition of the function
 ********************************************************************************/
/* LCD functions */
OzOLED LCD;

/*********************************************************************************
  SETUP

 ********************************************************************************/
void setup() {
  Serial.begin(115200);
#if !defined(__MIPSEL__)
  while (!Serial); // Wait for serial port to connect - used on Leonardo, Teensy and other boards with built-in USB CDC serial connection
#endif
  Serial.println("Start");
  if (Usb.Init() == -1)
    Serial.println("OSC did not start.");
  delay(200);

  if (!Hid.SetReportParser(0, &Joy))
    ErrorMessage<uint8_t > (PSTR("SetReportParser"), 1);
    
  /* S88 Control*/
      pinMode(2, INPUT_PULLUP);
      attachInterrupt(0,clk,RISING);    //pin 2 = clock interrupt
      pinMode(3, INPUT_PULLUP);
      attachInterrupt(1,PS,RISING);     //pin 3 = PS interrupt
      pinMode(dataIn,INPUT_PULLUP);     // data in from next Arduino S88 in chain
      pinMode(dataOut, OUTPUT);         // data out to commnad station or to previous Arduino in S88 chain
      digitalWrite(dataOut, LOW);   

  /* Loco Address setting LED */
      pinMode(PWR_LED_PIN, OUTPUT);
  /* Connmand Station Power on LED */
      pinMode(RUN_LED_PIN, OUTPUT);
      
  /* OLED init */
      LCD.OutputDammy = outputDammyPulse;
      LCD.init();
      LCD.clearDisplay();               //clear the screen and set start position to top left corner
      LCD.setNormalDisplay();           //Set display to Normal mode
      LCD.setPageMode();                //Set addressing mode to Page Mode
      LCD.printString("USB Mascon S88", 0, 2);
      LCD.printString("Ver1.03", 0, 3);
      LCD.printString("Nagoden", 0, 5);
      delay(2000);

  //EEPROMからアドレス読み出し
      EEPROM.get(0x000, glocoadress);
      if ((glocoadress > 0x270F) || (glocoadress == 0 ) ) {
          //EEPROMの値が9999より大きいか０の場合はアドレスを初期値３にする
          glocoadress = 3 ;
      }
      //Serial.print("EEPROM_R:");
      //Serial.println(glocoadress);
      gdecLocoadddress = glocoadress ;
 
  //表示用配列にセット
      disassemblyLocoaddress();

  //OLED初期表示
      clearDisplay() ;
      locoadddisp();
      dispMessageFirst();
  //方向表示
      Direction = 0;   // FWD
      dispDirection( Direction );
  //スピードバーの表示
      drawSpeedBar ( 9, 3, 0X00, 0);
  //画面
      SetAddress_mode = 0;

  //Reset task
      gPreviousL1 = millis();
      gPreviousL2 = millis();
      gPreviousL3 = millis();
}

/*********************************************************************************
  Loop処理
 ********************************************************************************/

void loop() {

  Usb.Task();


  /***********************************
    タスクスケジュール処理
    マスコンの位置で加減速処理
  ***********************************/

  if ( (millis() - gPreviousL1) >= 300)  {
      //Reset task
      gPreviousL1 = millis();

      /***********************************
        マスコンの位置で加減速処理
        加減速の調整はこの中のgSpeedの足し引き
        している値を調整可能
      ***********************************/
      //マスコンハンドルの文字表示
      switch (buttonBufRz)
      {
        case 0:  //EB
          gSpeed = gSpeed - CONST_EB ;
          break;
        case 0x05:  //B8
          gSpeed = gSpeed - CONST_B8 ;
          break;
        case 0x13:  //B7
          gSpeed = gSpeed - CONST_B7 ;
          break;
        case 0x20:  //B6
          gSpeed = gSpeed - CONST_B6 ;
          break;
        case 0x2E:  //B5
          gSpeed = gSpeed - CONST_B5 ;
          break;
        case 0x3C:  //B4
          gSpeed = gSpeed - CONST_B4 ;
          break;
        case 0x49:  //B3
          gSpeed = gSpeed - CONST_B3 ;
          break;
        case 0x57:  //B2
          gSpeed = gSpeed - CONST_B2 ;
          break;
        case 0x65:  //B1
          gSpeed = gSpeed - CONST_B1 ;
          break;
        case 0x80:  //N
          break;
        case 0x9F:  //P1
          if ((( gSpeed * 100) / 1023 ) < LIM_P1 ) {
            gSpeed = gSpeed + CONST_P1 ;      
          }
          break;
        case 0xB7:  //P2
          if ((( gSpeed * 100) / 1023 ) < LIM_P2 ) { 
            gSpeed = gSpeed + CONST_P2 ;
          }
          break;
        case 0xCE:  //P3
          if ((( gSpeed * 100) / 1023 ) < LIM_P3 ) { 
            gSpeed = gSpeed + CONST_P3;
          }
          break;
        case 0xE6:  //P4
          if ((( gSpeed * 100) / 1023 ) < LIM_P4 ) {
            gSpeed = gSpeed + CONST_P4;
          }
          break;
        case 0xFF:  //P5
          if ((( gSpeed * 100) / 1023 ) <= LIM_P5 ) {
            gSpeed = gSpeed + CONST_P5;
          }
          break;
      }

      if ( gSpeed <= 0  ) {
        gSpeed = 0 ;
      }
      if ( gSpeed > 1023 ) {
        gSpeed = 1023;
      }
      SpeedRate =  gSpeed / 4;
      //スピードバーの表示
      drawSpeedBar ( 9, 3, SpeedRate, 0);
      sprintf(gText, " %3d%c  ", ((word)( SpeedRate * 100) / 255) , 0x25 );
      printStringNOT ( gText , 9, 4);
      //スピードコマンド出力
      sendSpeed(SpeedRate);
  }

  /***********************************
    マスコンの位置がNの場合の減速処理
  ***********************************/
  if ( (millis() - gPreviousL2) >= 500)
  {
    //Reset task
      gPreviousL2 = millis();
    //マスコンハンドルがNの時で走行状態の場合に少しずつ減速
    if ((buttonBufRz == 0x80) && (gSpeed > 0) ) {
      gSpeed = gSpeed- CONST_NN ;
      drawSpeedBar ( 9, 3, SpeedRate, 0);   //スピードバーの表示
      sprintf(gText, " %3d%c  ", ((word)( SpeedRate * 100) / 255) , 0x25 );
      printStringNOT ( gText , 9, 4);
      sendSpeed(SpeedRate);                 //スピードコマンド出力
    }
  }

  /***********************************
    S88 バイトデータセット処理
  ***********************************/
  if ( (millis() - gPreviousL3) >= 100)
  {
    //Reset task
     gPreviousL3 = millis();
    //
      setS88_dir_speed();
  }

  /***********************************
    アイドル状態の時のキー処理
  ***********************************/
  //SetAddress_mode = 0 運転モード（何もボタン処理がない）
  //SetAddress_mode = 1 ZRボタンが押されアドレス変更モード
  //SetAddress_mode = 5 アクセサリ（ポイント）用のコマンド送出モード
  
  if (SetAddress_mode == 0 ) {
    //ZRボタンでアドレス設定画面に変移
    if ((buttonBufX1 & 0x80) == 0x80 ) {    //ZRボタン
      buttonBufX1 = 0x00;
      SetAddress_mode = 1 ;                 //アドレス変更モードに設定
      dispMessageAddset();                  //アドレス設定用のメッセージ
      dispMessageMode(1);                   //モード表示　AD
      //カーソルの反転
      wXcursor = 0;
      printStringNOT("  ", wXcursor , 4);   //アドレス番号の下にカーソル表示
      //現在のアドレス保持
      for (int i = 0; i < 4; i++) {
        gbkLocoAddresses[i] = gLocoAddresses[i];
      }
    }
    /***********************************
      方向反転の処理

    ***********************************/
    if ((buttonBufX2) == 0x02 ) {      //→ボタン
      buttonBufX2 = 0x0f;
      delay(100);
      Direction = 0;   // FWD
      dispDirection( Direction );
      loco_dir = Direction;
    }
    if ((buttonBufX2) == 0x06 ) {      //←ボタン
      buttonBufX2 = 0x0f;
      delay(100);
      Direction = 1;   // RWD
      dispDirection( Direction );
      loco_dir = Direction;
    }
    /***********************************
      ファンクションキーの処理

    ***********************************/
    setFunction() ;

  }

  /***********************************
    アドレス設定処理
    　  *SetAddress_mode == 1の時のキー処理
  ***********************************/
  if (SetAddress_mode == 1 ) {
      if ((buttonBufX1 & 0x02) == 0x02 ) {      //Bボタン
          buttonBufX1 = 0x00;
          dispMessageClear() ;                  //メッセージクリア
          dispMessageMode(0);                   //モード表示　クリア
          delay(100);
          dispMessageFirst();                   //初期メッセージ表示
          SetAddress_mode = 0 ;                 //運転モード
          cursorClear();                        //カーソルの反転クリア
          //アドレス保持の復帰
          for (int i = 0; i < 4; i++) {
              gLocoAddresses[i] = gbkLocoAddresses[i];
              sprintf(gText, "%1d", gLocoAddresses[i]);
              printStringBig(gText , i * 2, 0);
          }
      }
      if ((buttonBufX1 & 0x04) == 0x04 ) {      //Aボタン
          buttonBufX1 = 0x00;
          dispMessageClear() ;                  //メッセージクリア
          SetAddress_mode = 0 ;                 //運転モード
          //アドレスを変数にセット(10進数)
          int w;
          w = ( gLocoAddresses[0] * 1000 ) + ( gLocoAddresses[1] * 100 ) + ( gLocoAddresses[2] * 10) + gLocoAddresses[3];
          //            Serial.println( w);
          glocoadress = w;
          gdecLocoadddress = w;
         /*********************************************
             ここでアドレスをコマンドステーションに送る処理
             とアドレスを記憶する処理を記述(ディスパッチ処理）
          *********************************************/
           SetAddress_mode = 0;
           loco_no = glocoadress ;
           S88_type4_set_function();              //Locoアドレスの設定
           dispMessageClear();
           cursorClear();                         //カーソルの反転クリア
           dispMessageMode(0);                    //モード表示　クリア
           dispMessageFirst();
           AddressEEPROM_Write();
           digitalWrite(PWR_LED_PIN,HIGH);

      }
      //アドレス設定モードでさらにDRが押された時
      if ((buttonBufX1 & 0x80) == 0x80 ) {        //ZRボタン
          buttonBufX1 = 0x00;
          dispMessageswAddset();                  //アクセサリ用メッセージ表示
          dispMessageMode(2);                     //モード表示　SW
          swadddisp();                            //アクセサリ用アドレス表示
          SetAddress_mode = 5 ;                   //クセサリ（ポイント）用のコマンド送出モード
          //カーソルの反転
          wadXcursor = 0;
          printStringNOT("  ", wadXcursor , 4);
          //現在のアドレス保持
          for (int i = 0; i < 4; i++) {
            gbkswAddresses[i] = gswAddresses[i];
          }
   
      }
  }
    /************************************************　
      アドレス設定のキー処理
      矢印キーとA・Bボタンでアドレス決定
      ←→キーで桁のカーソル移動
      ↑↓キーで数値のUp/Down
      Aボタンで決定　Bボタンでキャンセル
    *************************************************/
    if ((buttonBufX2 == 0x02) && (SetAddress_mode == 1 )) {      //→ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        printString("  ", wXcursor * 2 , 4);
        wXcursor = wXcursor + 1;
        if ( wXcursor == 4 ) {
          wXcursor = 0;
        }
        printStringNOT("  ",  wXcursor * 2 , 4);
    }
    if ((buttonBufX2 == 0x06) && (SetAddress_mode == 1 )) {      //←ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        printString("  ", wXcursor * 2 , 4);
        wXcursor = wXcursor - 1;
        if ( wXcursor < 0 ) {
          wXcursor = 3;
        }
        printStringNOT("  ",  wXcursor * 2 , 4);
    }
    if ((buttonBufX2 == 0x00) && (SetAddress_mode == 1 )) {      //↑ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        gLocoAddresses[ wXcursor ] =  gLocoAddresses[ wXcursor ] + 1;
        if (gLocoAddresses[ wXcursor ] >= 10 ) {
           gLocoAddresses[ wXcursor ] = 0;
        }
        sprintf(gText, "%1d", gLocoAddresses[wXcursor]);
        printStringBig(gText , wXcursor * 2 , 0);
    }
    if ((buttonBufX2 == 0x04) && (SetAddress_mode == 1 )) {      //↓ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        if (gLocoAddresses[ wXcursor ] == 0 ) {
          gLocoAddresses[ wXcursor ] = 9;
        }
        else {
          gLocoAddresses[ wXcursor ] =  gLocoAddresses[ wXcursor ] - 1;
        }
        sprintf(gText, "%1d", gLocoAddresses[wXcursor]);
        printStringBig(gText , wXcursor * 2 , 0);
    }



  /*******************************************
    アクセサリアドレス設定
    　  *SetAddress_mode == 5の時のキー処理
  ********************************************/
    if (SetAddress_mode == 5 ) {
        if ((buttonBufX1 & 0x02) == 0x02 ) {      //Bボタン
          buttonBufX1 = 0x00;
          locoadddisp();
          cursorClear();    //カーソルの反転クリア
          dispMessageMode(0);
          dispMessageAddset();
          dispMessageFirst();
          SetAddress_mode = 0;
        }
        if ((buttonBufX1 & 0x04) == 0x04 ) {      //Aボタン（CLOSED）
          buttonBufX1 = 0x00;
          //アドレスを変数にセット(10進数)
          int w;
          w = ( gswAddresses[0] * 1000 ) + ( gswAddresses[1] * 100 ) + ( gswAddresses[2] * 10) + gswAddresses[3];
          glocoadress = w;
          acc_adr = w;
          acc_pos = 0;
          S88_type3_set_function();
        }
        if ((buttonBufX1 & 0x01) == 0x01 ) {      //Yボタン（THROWN）
          buttonBufX1 = 0x00;
          //アドレスを変数にセット(10進数)
          int w;
          w = ( gswAddresses[0] * 1000 ) + ( gswAddresses[1] * 100 ) + ( gswAddresses[2] * 10) + gswAddresses[3];
          glocoadress = w;
          acc_adr = w;
          acc_pos = 1;
          S88_type3_set_function();
        }
        if ((buttonBufX2) == 0x02 ) {             //→ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            printString("  ", wadXcursor * 2 , 4);
            wadXcursor = wadXcursor + 1;
            if ( wadXcursor == 4 ) {
                wadXcursor = 0;
            }
            printStringNOT("  ",  wadXcursor * 2 , 4);
        }
        if ((buttonBufX2) == 0x06 ) {             //←ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            printString("  ", wadXcursor * 2 , 4);
            wadXcursor = wadXcursor - 1;
            if ( wadXcursor < 0 ) {
                wadXcursor = 3;
            }
            printStringNOT("  ",  wadXcursor * 2 , 4);
        }
        if ((buttonBufX2 == 0x00 )&& (SetAddress_mode == 5 )) {      //↑ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            gswAddresses[ wadXcursor ] =  gswAddresses[ wadXcursor ] + 1;
            if (gswAddresses[ wadXcursor ] >= 10 ) {
              gswAddresses[ wadXcursor ] = 0;
            }
            sprintf(gText, "%1d", gswAddresses[wadXcursor]);
            printStringBig(gText , wadXcursor * 2 , 0);
        }
        if ((buttonBufX2 == 0x04) && (SetAddress_mode == 5 )) {      //↓ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            if (gswAddresses[ wadXcursor ] == 0 ) {
                gswAddresses[ wadXcursor ] = 9;
            }
            else {
              gswAddresses[ wadXcursor ] =  gswAddresses[ wadXcursor ] - 1;
            }
            sprintf(gText, "%1d", gswAddresses[wadXcursor]);
            printStringBig(gText , wadXcursor * 2 , 0);
        }
        
    }
}

/***********************************************************************
  各種関数
************************************************************************/
//EEPROMへのアドレス登録
void AddressEEPROM_Write (void) {
    EEPROM.put(0x000, glocoadress);
//  Serial.print("EEPROM_W:");
//  Serial.println(glocoadress);
}

// アドレスデータを４つの桁を表示用の配列にセット
void disassemblyLocoaddress(void){
    int work_Address = 0;
    int w;
    work_Address = gdecLocoadddress;
    gLocoAddresses[0] = work_Address / 1000 ;
    work_Address = work_Address - ((gLocoAddresses[0]) * 1000);
    gLocoAddresses[1] = work_Address / 100 ;
    work_Address = work_Address - (gLocoAddresses[1] * 100);
    gLocoAddresses[2] = work_Address / 10 ;
    work_Address = work_Address - (gLocoAddresses[2] * 10);
    gLocoAddresses[3] = work_Address;
}

//表示用の配列データをアドレスデータに変換
void assemblyLocoaddress(void){
    int work_Address;
    work_Address = ( gLocoAddresses[0] * 1000 ) + ( gLocoAddresses[1] * 100 ) + ( gLocoAddresses[2] * 10) + gLocoAddresses[3];
    //  Serial.println(work_Address);
    gdecLocoadddress = work_Address ;
}
//locoアドレス初期表示
void locoadddisp(void){
    //アドレス4桁表示
      for (int i = 0; i < 4; i++) {
        sprintf(gText, "%1d", gLocoAddresses[i]);
      printStringBig(gText , i * 2, 0);
      }
}
//アクセサリSWアドレス初期表示
void swadddisp(void){
    //アドレス3桁表示
      for (int i = 0; i < 4; i++) {
        sprintf(gText, "%1d", gswAddresses[i]);
      printStringBig(gText , (i * 2), 0);
      }
}

//進行方向表示　
void dispDirection(int n ) {
///////    Throttle.setDirection(n);
    if (n == 0) {
      //FWD表示
      printString ( "  FWD  " , 9, 2);
      drawIcon (14, 2, 2, 0);
      drawIcon (15, 2, 2, 0);
    }
    else {
      //RWD表示
      printString ( "  RWD  " , 9, 2);
      drawIcon (9, 2, 3, 0);
      drawIcon (10, 2, 3, 0);
    }
}

//OLEDのメッセージ枠のクリア
void dispMessageClear(void) {
    printString(space14 , 0, 5);
    printString(space14 , 0, 6);
    printString(space14, 0, 7);
}
//OLEDのメッセージ 初期画面
void dispMessageFirst(void) {
    printString(space14 , 0, 5);
    printString("ZR:Set LocoAdr", 0, 6);
    printString("<>:Direction  ", 0, 7);
}
//OLEDのメッセージ アドレス設定画面
void dispMessageAddset(void) {
    printString("<>UpDn: Select", 0, 5);
    printString("A: Decision   ", 0, 6);
    printString("B: Exit       ", 0, 7);
}
//OLEDのメッセージ スイッチアドレス設定画面
void dispMessageswAddset(void) {
    printString("<>UpDn: Select", 0, 5);
    printString("A-> c  Y-> t  ", 0, 6);
    printString("B: Exit    ", 0, 7);
}
//OLEDのメッセージ アドレス設定　Steal確認画面
void dispMessageSteal(void) {
    printString("Steal y/n? ", 0, 5);
    printString("A: Yes     ", 0, 6);
    printString("B: No      ", 0, 7);
}
//OLEDのメッセージ モード画面
void dispMessageMode(int dmode) {
    switch(dmode){
      case 0:      //clear
          LCD.printString("  ", 9, 0);
          LCD.printString("  ", 9, 1);
          break; 
      case 1:      //LocoAddress
          LCD.printStringNOT("AD", 9, 0);
          LCD.printString("  ", 9, 1);
          break; 
      case 2:      //AccessoryAddress
          LCD.printString("  ", 9, 0);
          LCD.printStringNOT("SW", 9, 1);
          break; 
    }
}
//OLEDのメッセージ アドレス表示バーのカーソルクリア
void cursorClear(void){
      printString("        ", 0, 4);           //カーソルの反転クリア
}
//スロットルスピードコマンド処理
void sendSpeed(int spd) {
    //減速停止時にスピードが０にならないための対応処理
    if ((gMspeed > 0 ) && ( spd == 0 )){
        gMspeed = 0;
    }
    else {
        spd = spd / 4;          //本来ならスピードステップに合わせて計算すべき
        gMspeed = spd;          //スピード値の保持
    }

    //スピードコマンド出力
    loco_speed = gMspeed  ;
}

//スピードステップの設定
void SendStepmode( int Stepmode){

}
/*****************************************************************************
//マスコンの押しボタンでファンクション送信
//ファンクションと押しボタンとのマッピングはここで変更してね
*****************************************************************************/
void setFunction(void) {
  //マスコンボタン処理その１
  if ((buttonBufX1 & 0x04) == 0x04 ) {    //ボタンA
    if (BUTTON_A != 99) {
        ValueFunction[BUTTON_A] = !ValueFunction[BUTTON_A];        //ボタンA
    }    
  }
  if ((buttonBufX1 & 0x02) == 0x02 ) {    //ボタンB
    if (BUTTON_B != 99) {
        ValueFunction[BUTTON_B] = !ValueFunction[BUTTON_B];        //ボタンB
    }
  }
  if ((buttonBufX1 & 0x01) == 0x01 ) {    //ボタンY
    if (BUTTON_Y != 99) {
        ValueFunction[BUTTON_Y] = !ValueFunction[BUTTON_Y];        //ボタンY
    }    
  }
  if ((buttonBufX1 & 0x08) == 0x08 ) {    //ボタンX
    if (BUTTON_X != 99) {
        ValueFunction[BUTTON_X] = !ValueFunction[BUTTON_X];        //ボタンX
    }    
  }
  if ((buttonBufX1 & 0x20) == 0x20 ) {    //ボタンR
    if (BUTTON_R != 99) {
        ValueFunction[BUTTON_R] = !ValueFunction[BUTTON_R];        //ボタンX
    }    
  }
  if ((buttonBufX1 & 0x10) == 0x10 ) {    //ボタンL
    if (BUTTON_L != 99) {
        ValueFunction[BUTTON_L] = !ValueFunction[BUTTON_L];        //ボタンL
    }    
  }
  if ((buttonBufX1 & 0x40) == 0x40 ) {    //ボタンZL
    if (BUTTON_ZL != 99) {
        ValueFunction[BUTTON_ZL] = !ValueFunction[BUTTON_ZL];      //ボタンZL
    }    
    //マスコンのEBを入れるとZLボタンが同時にセットされる仕様であることに注意
  }
  if ((buttonBufY1 & 0x20) == 0x20 ) {                             //ボタン● 線路電源ON/OFF
      Power_Flg = !Power_Flg;
      S88_type4_set_Power(Power_Flg);
      if (Power_Flg == 1){
       digitalWrite(RUN_LED_PIN, HIGH);
      } else {
       digitalWrite(RUN_LED_PIN, LOW);
      }
  }
  if ((buttonBufY1 & 0x10) == 0x10 ) {                             //ボタンHOME
      //Homeボタンは緊急停止（EBの機能とは違ってＤＣＣでの緊急停止です）
        gSpeed = 0;
      //緊急停止コマンド送出をここに入れる
  }
  if ((buttonBufY1 & 0x02) == 0x02 ) {                             //ボタン+
    if (BUTTON_P != 99) {
        ValueFunction[BUTTON_P] = !ValueFunction[BUTTON_P];       
    }    
  }
  if ((buttonBufY1 & 0x01) == 0x01 ) {                             //ボタン-
    if (BUTTON_M != 99) {
        ValueFunction[BUTTON_M] = !ValueFunction[BUTTON_M];       
    }    
  }
  buttonBufX1 = 0x00;
  buttonBufY1 = 0x00;

  //      ZRはアドレス設定用(使用禁止)
  //      if ((buttonBufX1 & 0x80) == 0x80 ) {    //ボタンZR
  //      }
}

/***********************************************************************
  OLEDの関数
************************************************************************/
void printString(const char *s, int x, int y)
{
  LCD.printString(s, x, y);
}

void printStringNOT(const char *s, int x, int y)
{
  LCD.printStringNOT(s, x, y);
}

void printStringBig(const char *s, int x, int y)
{
  LCD.printBigNumberX2(s, x, y);
}

void drawIcon(byte x, byte y, byte num, byte inNot)
{
  LCD.drawIcon(x, y, num, inNot);
}

void drawSpeedBar(byte x, byte y, byte speed, byte inNot)
{
  LCD.drawSpeedBar(x, y, speed, false);
}

void clearDisplay()
{
  LCD.lcd_clear();           //clear the screen and set start position to top left corner
}
void outputDammyPulse()
{
}
/*************************************************************************************************
  USB　Host Sheildの受信イベント
 *************************************************************************************************/
void MyJoystickEvents::OnGamePadChanged(const GamePadEventData *evt) {
/*
          Serial.print("X1: ");
          Serial.print(evt->X , HEX);
          Serial.print("\tY1: ");
          Serial.print(evt->Y , HEX);
          Serial.print("\tX2: ");
          Serial.print(evt->Z1 , HEX);
          Serial.print("\tY2: ");
          Serial.print(evt->Z2 , HEX);
          Serial.print("\tRz: ");
          Serial.print(evt->Rz , BIN);
          Serial.println("");
*/  
  buttonBufRz = evt->Rz ;
  buttonBufX1 = evt->X ;
  buttonBufX2 = evt->Z1 ;
  buttonBufY1 = evt->Y;

  //マスコンハンドルの文字表示
  switch (buttonBufRz)
  {
    case 0:
      sprintf(gText, "EB");
      LCD.printStringNOT(gText, 14, 0);
      break;
    case 0x05:
      sprintf(gText, "B8");
      LCD.printString(gText, 14, 0);
      break;
    case 0x13:
      sprintf(gText, "B7");
      LCD.printString(gText, 14, 0);
      break;
    case 0x20:
      sprintf(gText, "B6");
      LCD.printString(gText, 14, 0);
      break;
    case 0x2E:
      sprintf(gText, "B5");
      LCD.printString(gText, 14, 0);
      break;
    case 0x3C:
      sprintf(gText, "B4");
      LCD.printString(gText, 14, 0);
      break;
    case 0x49:
      sprintf(gText, "B3");
      LCD.printString(gText, 14, 0);
      break;
    case 0x57:
      sprintf(gText, "B2");
      LCD.printString(gText, 14, 0);
      break;
    case 0x65:
      sprintf(gText, "B1");
      LCD.printString(gText, 14, 0);
      break;
    case 0x80:
      sprintf(gText, "N ");
      LCD.printString(gText, 14, 0);
      break;
    case 0x9F:
      sprintf(gText, "P1");
      LCD.printString(gText, 14, 0);
      break;
    case 0xB7:
      sprintf(gText, "P2");
      LCD.printString(gText, 14, 0);
      break;
    case 0xCE:
      sprintf(gText, "P3");
      LCD.printString(gText, 14, 0);
      break;
    case 0xE6:
      sprintf(gText, "P4");
      LCD.printString(gText, 14, 0);
      break;
    case 0xFF:
      sprintf(gText, "P5");
      LCD.printString(gText, 14, 0);
      break;
  }
  /********************************************************
     以下このイベント内で使用しないのでコメントアウトにする


   ********************************************************/
  /*
          //マスコンボタン処理その１
          if ((buttonBufX1 & 0x04) == 0x04 ) {         //ボタンA
          }
          if ((buttonBufX1 & 0x02) == 0x02 ) {         //ボタンB
          }
          if ((buttonBufX1 & 0x01) == 0x01 ) {         //ボタンY
          }
          if ((buttonBufX1 & 0x08) == 0x08 ) {         //ボタンX
          }
          if ((buttonBufX1 & 0x20) == 0x20 ) {         //ボタンR
          }
          if ((buttonBufX1 & 0x10) == 0x10 ) {         //ボタンL
          }
          if ((buttonBufX1 & 0x40) == 0x40 ) {         //ボタンZL
          }
          if ((buttonBufX1 & 0x80) == 0x80 ) {         //ボタンZR
          }
          //マスコンボタン処理その２（矢印キー）
          switch(buttonBufX2)
          {
           case 0x00:        //ボタン↑
              break;
           case 0x02:        //ボタン→
              break;
           case 0x04:        //ボタン↓
              break;
           case 0x06:        //ボタン←
              break;
           //default:
          }
  */

}

//S88
void PS()
{
  //Set transfer value
  clk_counter=0;
  out_flag = true;
  out_value = sensor_value;
  //0リセットはボタン系だけ
  //sensor_value &= 0x07FF;
}

//S88
void clk()
{
  //transfer 
  if(out_flag == true)
  {
    digitalWrite(dataOut,(out_value >> clk_counter) & 0x1);
  }
  else
  {
    digitalWrite(dataOut,(transfer_value >> clk_counter) & 0x1);
  }
  //Delay makes reading output signal from next Arduino in chain more reliable.
  delayMicroseconds(16); 
  //Store dataIn value
  //transfer_value[clk_counter] = digitalRead(dataIn);
  if(digitalRead(dataIn) == 1)
  {
     transfer_value |= 1 << clk_counter;
  }
  else
  {
     transfer_value &= ~(1 << clk_counter);
  }

  //  clk_counter =(clk_counter +1) % CONT_PORT_NUM;
  if(clk_counter < CONT_PORT_NUM - 1)
  {
    clk_counter ++;
  }
  else
  {
    clk_counter = 0;
    out_flag = false;
    SendReserveFlag = 100;//SendFlagリセット
  }
}

uint16_t parity_calc(uint16_t inData)
{
  uint16_t aParity = 0;
   /* パリティ計算 */
  for( int i = 0; i < 15; i++)
  {
    aParity = aParity ^ ((inData >> i) & 1);
  }
  return aParity;
}

void S88_type0_set_function()
{
  //Type0 全書き
  Serial.println("S88_type0_set_function: "); 
  SendReserveFlag = 0;
//  uint16_t speed63 =(((uint16_t) loco_speed) >> 5) & 0x003F;
  uint16_t speed63 =(loco_speed & 0x003F);
  sensor_value = 0x0000;
  sensor_value |= speed63;
  sensor_value |= (uint16_t)loco_dir << 6;
  for(int i = 0;i < 5;i++)
  {
        sensor_value |= ((uint16_t)ValueFunction[i]) << (7 + i);
//      sensor_value |= ((uint16_t)ValueFunction[i] & 0b1) << (7 + i);
  }
  sensor_value |= parity_calc(sensor_value) << 15;  
}

void S88_type1_set_function()
{
  //Type1 fn5-16全書き
  Serial.println("S88_type1_set_function"); 
  SendReserveFlag = 1;
  sensor_value = 0x0000;
  sensor_value |= 0x1000;
  for(int i = 0;i < 12;i++)
  {
    //sensor_value |= ((uint16_t)ValueFunction[i+5]) << i;
      sensor_value |= (ValueFunction[i+5] & 0b1) << i;
  }
  sensor_value |= parity_calc(sensor_value) << 15;  
}

void S88_type2_set_function()
{
  //Type2 fn17-28 全書き
  Serial.println("S88_type2_set_function"); 
  SendReserveFlag = 2;
  sensor_value = 0x0000;
  sensor_value |= 0x2000;
  for(int i = 0;i < 12;i++)
  {
    //sensor_value |= ((uint16_t)ValueFunction[i+17]) << i;
    sensor_value |= ((uint16_t)ValueFunction[i+17] & 0b1) << i;
  }
  sensor_value |= parity_calc(sensor_value) << 15;  
}
void S88_type3_set_function()
{
  //Type3 ACC Control
  Serial.println("S88_type3_set_function");
  if ((acc_adr > 0) && (acc_adr <2045)){   
    SendReserveFlag = 3;
    uint16_t acaddr =(acc_adr & 0x07FF);
    sensor_value = 0x0000;
    sensor_value |= 0x3000;
    sensor_value |= acaddr;
    sensor_value |= (uint16_t)acc_pos << 11;
    sensor_value |= parity_calc(sensor_value) << 15;
  } else{
     Serial.print("ACC Address value is out of range:");
     Serial.println(acc_adr);
  }
}
void S88_type4_set_function()
{
  //Type4 LocoNo 全書き
  Serial.println("S88_type4_set_function "); 
  SendReserveFlag = 4;
  sensor_value = 0x0000;
  sensor_value |= 0x4000;
  sensor_value |= (uint16_t)loco_no;
  sensor_value |= parity_calc(sensor_value) << 15;  

}
void S88_type4_set_Power( boolean Power_SW)
{
  //Type4 Power On off
  Serial.println("S88_type4_set_Power"); 
  SendReserveFlag = 4;
  sensor_value = 0x0000;
  sensor_value |= 0x4000;
  if (Power_SW ){
      sensor_value |= (uint16_t)16382;
  } else{
      sensor_value |= (uint16_t)0;
  }
  sensor_value |= parity_calc(sensor_value) << 15;  

}

void setS88_dir_speed()
{
  //direction,speed
  //定期送信時はLocAddress,その他を順番に送信する。
  if(SendReserveFlag == 100)
  {
    if((loco_speed != loco_speed_prev) || (loco_dir != loco_dir_prev))
    {
      //スピードが変更なら
      //direction,speed
      S88_type0_set_function();
      Serial.println("dir_speed"); 
    }
    else
    {
      //定期送信
      switch(set_s88_switch)
      {
        case 0:
          S88_type0_set_function();
          break;
        case 1:
          S88_type1_set_function();
          break;
        case 2:
          S88_type2_set_function();
          break;
        case 4:
        default:
          S88_type4_set_function();
          break;
      }
      //トグル
      if(set_s88_switch == 4)
      {
        set_s88_switch = 0;
      }
      else
      {
        set_s88_switch ++;
      }
    }
  }
  loco_speed_prev = loco_speed;
  loco_dir_prev = loco_dir;
}

void setS88_function(int change_functin_number)
{
  if(change_functin_number <= 4)
  {
    S88_type0_set_function();
  }
  else if(change_functin_number <= 16)
  {
    S88_type1_set_function();
  }
  else //17 - 28
  {
    S88_type2_set_function();
  }
}
